"""Converter classes for builtin types."""

from . import common
from .base import Converter

log = common.logger(__name__)


class Object(Converter):  # pylint: disable=W0223

    """Base class for standard types (mapped directly to YAML)."""

    @classmethod
    def to_value(cls, obj):
        return obj

    @classmethod
    def to_data(cls, obj):
        return cls.to_value(obj)


class String(Object):

    """Converter for the `str` type."""

    TYPE = str
    DEFAULT = ""

    @classmethod
    def to_value(cls, obj):
        if isinstance(obj, cls.TYPE):
            return obj
        elif obj:
            try:
                return ', '.join(str(item) for item in obj)
            except TypeError:
                return str(obj)
        else:
            return cls.DEFAULT


class Integer(Object):

    """Converter for the `int` type."""

    TYPE = int
    DEFAULT = 0

    @classmethod
    def to_value(cls, obj):
        if isinstance(obj, cls.TYPE):
            return obj
        elif obj:
            try:
                return int(obj)
            except ValueError:
                return int(float(obj))
        else:
            return cls.DEFAULT


class Float(Object):

    """Converter for the `float` type."""

    TYPE = float
    DEFAULT = 0.0

    @classmethod
    def to_value(cls, obj):
        if isinstance(obj, cls.TYPE):
            return obj
        elif obj:
            return float(obj)
        else:
            return cls.DEFAULT


class Boolean(Object):

    """Converter for the `bool` type."""

    TYPE = bool
    DEFAULT = False

    FALSY = ('false', 'f', 'no', 'n', 'disabled', 'off', '0')

    @classmethod
    def to_value(cls, obj):
        if isinstance(obj, str) and obj.lower().strip() in cls.FALSY:
            return False
        elif obj is not None:
            return bool(obj)
        else:
            return cls.DEFAULT


def match(name, data, nested=False):
    """Determine the appropriate converter for new data."""
    nested = " nested" if nested else ""
    msg = "determining converter for new%s: '%s' = %r"
    log.debug(msg, nested, name, repr(data))

    converters = Object.__subclasses__()
    log.trace("converter options: {}".format(converters))

    for converter in converters:
        if converter.TYPE and type(data) == converter.TYPE:
            log.debug("matched converter: %s", converter)
            log.info("new%s attribute: %s", nested, name)
            return converter

    if data is None or isinstance(data, (dict, list)):
        log.info("default converter: %s", Object)
        log.warn("new%s attribute with unknown type: %s", nested, name)
        return Object

    raise common.ConversionError("no converter available for: {}".format(data))



#!/usr/bin/env python
# pylint:disable=R0201

"""Unit tests for the `container` module."""

import pytest

from yorm.utilities import attr
from yorm.container import Dictionary, List
from yorm.standard import String, Integer


# sample classes ##############################################################


@attr(abc=Integer)
class SampleDictionary(Dictionary):

    """Sample dictionary container."""


@attr(var1=Integer, var2=String)
class SampleDictionaryWithInitialization(Dictionary):

    """Sample dictionary container with initialization."""

    def __init__(self, var1, var2, var3):
        super().__init__()
        self.var1 = var1
        self.var2 = var2
        self.var3 = var3


@attr(all=String)
class StringList(List):

    """Sample list container."""

    # TODO: this shouldn't be required for the tests to pass
    yorm_attrs = {'all': String}


class UnknownList(List):

    """Sample list container."""


# tests #######################################################################


class TestDictionary:

    """Unit tests for the `Dictionary` container."""

    obj = {'abc': 123}

    data_value = [
        (obj, obj),
        (None, {'abc': 0}),
        ("key=value", {'key': "value", 'abc': 0}),
        ("key=", {'key': "", 'abc': 0}),
        ("key", {'key': None, 'abc': 0}),
    ]

    value_data = [
        (obj, obj),
    ]

    def setup_method(self, _):
        """Reset the class' mapped attributes before each test."""
        SampleDictionary.yorm_attrs = {'abc': Integer}

    @pytest.mark.parametrize("data,value", data_value)
    def test_to_value(self, data, value):
        """Verify input data is converted to values."""
        assert value == SampleDictionary.to_value(data)

    @pytest.mark.parametrize("value,data", value_data)
    def test_to_data(self, value, data):
        """Verify values are converted to output data."""
        assert data == SampleDictionary.to_data(value)

    def test_not_implemented(self):
        """Verify `Dictionary` cannot be used directly."""
        with pytest.raises(NotImplementedError):
            Dictionary.to_value(None)
        with pytest.raises(NotImplementedError):
            Dictionary.to_data(None)

    def test_dict_as_object(self):
        """Verify a `Dictionary` can be used as an attribute."""
        dictionary = SampleDictionaryWithInitialization(1, 2, 3.0)
        value = {'var1': 1, 'var2': '2'}
        value2 = dictionary.to_value(dictionary)
        assert value == value2
        # keys are not accesible as attributes
        assert not hasattr(value2, 'var1')
        assert not hasattr(value2, 'var2')
        assert not hasattr(value2, 'var3')


class TestList:

    """Unit tests for the `List` container."""

    obj = ["a", "b", "c"]

    data_value = [
        (obj, obj),
        (None, []),
        ("a b c", ["a", "b", "c"]),
        ("a,b,c", ["a", "b", "c"]),
        ("abc", ["abc"]),
        ("a\nb\nc", ["a", "b", "c"]),
        (4.2, ['4.2']),
    ]

    value_data = [
        (obj, obj),
    ]

    @pytest.mark.parametrize("data,value", data_value)
    def test_to_value(self, data, value):
        """Verify input data is converted to values."""
        assert value == StringList.to_value(data)

    @pytest.mark.parametrize("value,data", value_data)
    def test_to_data(self, value, data):
        """Verify values are converted to output data."""
        assert data == StringList.to_data(value)

    def test_item_type(self):
        """Verify list item type can be determined."""
        assert String == StringList.item_type

    def test_item_type_none(self):
        """Verify list item type defaults to None."""
        assert None is UnknownList.item_type

    def test_not_implemented(self):
        """Verify `List` cannot be used directly."""
        with pytest.raises(NotImplementedError):
            List.to_value(None)
        with pytest.raises(NotImplementedError):
            List.to_data(None)
        with pytest.raises(NotImplementedError):
            UnknownList.to_value(None)
        with pytest.raises(NotImplementedError):
            UnknownList.to_data(None)


if __name__ == '__main__':
    pytest.main()



#!/usr/bin/env python
# pylint:disable=R0201

"""Integration tests for the `yorm` package."""

import time
import logging

import pytest

from yorm import sync, attr, Converter
from yorm.container import Dictionary, List
from yorm.standard import Object, String, Integer, Float, Boolean
from yorm.extended import Markdown

integration = pytest.mark.integration


# custom converters ###########################################################


class Level(Converter):

    """Sample custom attribute."""

    @classmethod
    def to_value(cls, obj):
        if obj:
            if isinstance(obj, str):
                return obj
            else:
                return str(obj)
        else:
            return ""

    @classmethod
    def to_data(cls, obj):
        value = cls.to_value(obj)
        count = value.split('.')
        if count == 0:
            return int(value)
        elif count == 1:
            return float(value)
        else:
            return value


class EmptyDictionary(Dictionary):

    """Sample dictionary container."""


@attr(key=String)
class SingleKeyDictionary(Dictionary):

    """Sample dictionary container."""


@attr(all=Integer)
class IntegerList(List):

    """Sample list container."""


@attr(status=Boolean, label=String)
class StatusDictionary(Dictionary):

    """Sample dictionary container."""


@attr(all=StatusDictionary)
class StatusDictionaryList(List):

    """Sample list container."""


# sample classes ##############################################################


class SampleStandard:

    """Sample class using standard attribute types."""

    def __init__(self):
        # https://docs.python.org/3.4/library/json.html#json.JSONDecoder
        self.object = {}
        self.array = []
        self.string = ""
        self.number_int = 0
        self.number_real = 0.0
        self.true = True
        self.false = False
        self.null = None

    def __repr__(self):
        return "<standard {}>".format(id(self))


class SampleNested:

    """Sample class using nested attribute types."""

    def __init__(self):
        self.count = 0
        self.results = []

    def __repr__(self):
        return "<nested {}>".format(id(self))


@attr(object=EmptyDictionary, array=IntegerList, string=String)
@attr(number_int=Integer, number_real=Float)
@attr(true=Boolean, false=Boolean)
@sync("path/to/{self.category}/{self.name}.yml")
class SampleStandardDecorated:

    """Sample class using standard attribute types."""

    def __init__(self, name, category='default'):
        self.name = name
        self.category = category
        # https://docs.python.org/3.4/library/json.html#json.JSONDecoder
        self.object = {}
        self.array = []
        self.string = ""
        self.number_int = 0
        self.number_real = 0.0
        self.true = True
        self.false = False
        self.null = None

    def __repr__(self):
        return "<decorated {}>".format(id(self))


@attr(string=String, number_real=Float)
@sync("sample.yml", auto=False)
class SampleDecoratedNoAuto:

    """Sample class with automatic storage turned off."""

    def __init__(self):
        self.string = ""
        self.number_real = 0.0

    def __repr__(self):
        return "<no auto {}>".format(id(self))


@attr(string=String, number_real=Float)
class SampleDecoratedNoPath:

    """Sample class with a manually mapped path."""

    def __init__(self):
        self.string = ""
        self.number_real = 0.0

    def __repr__(self):
        return "<no path {}>".format(id(self))


@sync("sample.yml")
class SampleEmptyDecorated:

    """Sample class using standard attribute types."""

    def __repr__(self):
        return "<empty {}>".format(id(self))


class SampleExtended:

    """Sample class using extended attribute types."""

    def __init__(self):
        self.text = ""

    def __repr__(self):
        return "<extended {}>".format(id(self))


@sync("path/to/directory/{UUID}.yml", attrs={'level': Level})
class SampleCustomDecorated:

    """Sample class using custom attribute types."""

    def __init__(self, name):
        self.name = name
        self.level = '1.0'

    def __repr__(self):
        return "<custom {}>".format(id(self))


# tests #######################################################################


def refresh_file_modification_times(seconds=1.1):
    """Sleep to allow file modification times to refresh."""
    logging.info("delaying for %s second%s...", seconds,
                 "" if seconds == 1 else "s")
    time.sleep(seconds)


def test_imports():
    """Verify the package namespace is correct."""
    # pylint: disable=W0404,W0612,W0621

    import yorm

    # Constants
    from yorm import UUID  # filename placeholder

    # Classes
    from yorm import Mappable  # base class for mapped objects
    from yorm import Converter  # base class for converters
    from yorm.standard import String  # and others
    from yorm.extended import Markdown  # and others
    from yorm.container import List  # and others

    # Decorators
    from yorm import sync  # enables mapping on a class's instance objects
    from yorm import sync_instances  # alias for the class decorator
    from yorm import attr  # alternate API to identify mapped attributes

    # Functions
    from yorm import sync  # enables mapping on an instance object
    from yorm import sync_object  # alias for the mapping function
    from yorm import update  # fetch (if necessary) and store a mapped object
    from yorm import update_object  # fetch (optional force) a mapped object
    from yorm import update_file  # store a mapped object


@integration
class TestStandard:

    """Integration tests for standard attribute types."""

    def test_decorator(self, tmpdir):
        """Verify standard attribute types dump/load correctly (decorator)."""
        tmpdir.chdir()
        sample = SampleStandardDecorated('sample')
        assert "path/to/default/sample.yml" == sample.yorm_path

        # check defaults
        assert {} == sample.object
        assert [] == sample.array
        assert "" == sample.string
        assert 0 == sample.number_int
        assert 0.0 == sample.number_real
        assert True is sample.true
        assert False is sample.false
        assert None is sample.null

        # change object values
        sample.object = {'key2': 'value'}
        sample.array = [0, 1, 2]
        sample.string = "Hello, world!"
        sample.number_int = 42
        sample.number_real = 4.2
        sample.true = False
        sample.false = True

        # check file values
        with open(sample.yorm_path, 'r') as stream:
            text = stream.read()
        assert """
        array:
        - 0
        - 1
        - 2
        'false': true
        number_int: 42
        number_real: 4.2
        object:
          key2: value
        string: Hello, world!
        'true': false
        """.strip().replace("        ", "") + '\n' == text

        # change file values
        refresh_file_modification_times()
        text = """
        array: [4, 5, 6]
        'false': null
        number_int: 42
        number_real: '4.2'
        object: {'status': false}
        string: "abc"
        'true': null
        """.strip().replace("        ", "") + '\n'
        with open(sample.yorm_path, 'w') as stream:
            stream.write(text)

        # check object values
        assert {'key2': "",
                'status': False} == sample.object
        assert [4, 5, 6] == sample.array
        assert "abc" == sample.string
        assert 42 == sample.number_int
        assert 4.2 == sample.number_real
        assert False is sample.true
        assert False is sample.false

    def test_function(self, tmpdir):
        """Verify standard attribute types dump/load correctly (function)."""
        tmpdir.chdir()
        _sample = SampleStandard()
        attrs = {'object': SingleKeyDictionary,
                 'array': IntegerList,
                 'string': String,
                 'number_int': Integer,
                 'number_real': Float,
                 'true': Boolean,
                 'false': Boolean}
        sample = sync(_sample, "path/to/directory/sample.yml", attrs)
        assert "path/to/directory/sample.yml" == sample.yorm_path

        # check defaults
        assert {} == sample.object
        assert [] == sample.array
        assert "" == sample.string
        assert 0 == sample.number_int
        assert 0.0 == sample.number_real
        assert True is sample.true
        assert False is sample.false
        assert None is sample.null

        # change object values
        sample.object = {'key2': 'value'}
        sample.array = [1, 2, 3]
        sample.string = "Hello, world!"
        sample.number_int = 42
        sample.number_real = 4.2
        sample.true = None
        sample.false = 1

        # check file values
        with open(sample.yorm_path, 'r') as stream:
            text = stream.read()
        assert """
        array:
        - 1
        - 2
        - 3
        'false': true
        number_int: 42
        number_real: 4.2
        object:
          key: ''
          key2: value
        string: Hello, world!
        'true': false
        """.strip().replace("        ", "") + '\n' == text

    def test_with(self, tmpdir):
        """Verify standard attribute types dump/load correctly (with)."""
        tmpdir.chdir()
        _sample = SampleStandard()
        attrs = {'string': String,
                 'number_real': Float}
        sample = sync(_sample, "sample.yml", attrs, auto=False)

        # change object values
        with sample:
            sample.string = "abc"
            sample.number_real = 4.2

            # check for unchanged file values
            with open(sample.yorm_path, 'r') as stream:
                text = stream.read()
            assert "" == text

            # check for changed file values
        with open(sample.yorm_path, 'r') as stream:
            text = stream.read()
        assert """
        number_real: 4.2
        string: abc
        """.strip().replace("        ", "") + '\n' == text

    def test_no_auto(self, tmpdir):
        """Verify standard attribute types dump/load correctly (auto)."""
        tmpdir.chdir()
        sample = SampleDecoratedNoAuto()

        # change object values
        sample.string = "abc"
        sample.number_real = 4.2

        # check for unchanged file values
        with open(sample.yorm_path, 'r') as stream:
            text = stream.read()
        assert "" == text

        # store value
        sample.yorm_mapper.store(sample, sample.yorm_attrs)
        sample.yorm_mapper.auto = True

        # check for changed file values
        with open(sample.yorm_path, 'r') as stream:
            text = stream.read()
        assert """
        number_real: 4.2
        string: abc
        """.strip().replace("        ", "") + '\n' == text

    def test_no_path(self, tmpdir):
        """Verify standard attribute types dump/load correctly (path)."""
        tmpdir.chdir()
        sample = sync(SampleDecoratedNoPath(), "sample.yml")

        # change object values
        sample.string = "abc"
        sample.number_real = 4.2

        # check for changed file values
        with open(sample.yorm_path, 'r') as stream:
            text = stream.read()
        assert """
        number_real: 4.2
        string: abc
        """.strip().replace("        ", "") + '\n' == text


@integration
class TestContainers:

    """Integration tests for attribute containers."""

    def test_nesting(self, tmpdir):
        """Verify standard attribute types can be nested."""
        tmpdir.chdir()
        _sample = SampleNested()
        attrs = {'count': Integer,
                 'results': StatusDictionaryList}
        sample = sync(_sample, "sample.yml", attrs)

        # check defaults
        assert 0 == sample.count
        assert [] == sample.results

        # change object values
        sample.count = 5
        sample.results = [{'status': False, 'label': "abc"},
                          {'status': None, 'label': None},
                          {'label': "def"},
                          {'status': True},
                          {}]

        # check file values
        with open(sample.yorm_path, 'r') as stream:
            text = stream.read()
        assert """
        count: 5
        results:
        - label: abc
          status: false
        - label: ''
          status: false
        - label: def
          status: false
        - label: ''
          status: true
        - label: ''
          status: false
        """.strip().replace("        ", "") + '\n' == text

        # change file values
        refresh_file_modification_times()
        text = """
        count: 3
        other: 4.2
        results:
        - label: abc
        - label: null
          status: false
        - status: true
        """.strip().replace("        ", "") + '\n'
        with open(sample.yorm_path, 'w') as stream:
            stream.write(text)

        # check object values
        assert 3 == sample.count
        assert 4.2 == sample.other
        assert [{'label': 'abc', 'status': False},
                {'label': '', 'status': False},
                {'label': '', 'status': True}] == sample.results

    def test_objects(self, tmpdir):
        """Verify containers are treated as objects when added."""
        tmpdir.chdir()
        sample = SampleEmptyDecorated()

        # change file values
        refresh_file_modification_times()
        text = """
        object: {'key': 'value'}
        array: [1, '2', '3.0']
        """.strip().replace("        ", "") + '\n'
        with open(sample.yorm_path, 'w') as stream:
            stream.write(text)

        # (a mapped attribute must be read first to trigger retrieving)
        sample.yorm_mapper.fetch(sample, sample.yorm_attrs)

        # check object values
        assert {'key': 'value'} == sample.object
        assert [1, '2', '3.0'] == sample.array

        # check object types
        assert Object == sample.yorm_attrs['object']
        assert Object == sample.yorm_attrs['array']

        # change object values
        sample.object = None  # pylint: disable=W0201
        sample.array = "abc"  # pylint: disable=W0201

        # check file values
        with open(sample.yorm_path, 'r') as stream:
            text = stream.read()
        assert """
        array: abc
        object: null
        """.strip().replace("        ", "") + '\n' == text


@integration
class TestExtended:

    """Integration tests for extended attribute types."""

    def test_function(self, tmpdir):
        """Verify extended attribute types dump/load correctly."""
        tmpdir.chdir()
        _sample = SampleExtended()
        attrs = {'text': Markdown}
        sample = sync(_sample, "path/to/directory/sample.yml", attrs)

        # check defaults
        assert "" == sample.text

        # change object values
        refresh_file_modification_times()
        sample.text = """
        This is the first sentence. This is the second sentence.
        This is the third sentence.
        """.strip().replace("        ", "")

        # check file values
        with open(sample.yorm_path, 'r') as stream:
            text = stream.read()
        assert """
        text: |
          This is the first sentence.
          This is the second sentence.
          This is the third sentence.
        """.strip().replace("        ", "") + '\n' == text

        # change file values
        refresh_file_modification_times()
        text = """
        text: |
          This is a
          sentence.
        """.strip().replace("        ", "") + '\n'
        with open(sample.yorm_path, 'w') as stream:
            stream.write(text)

        # check object values
        assert "This is a sentence." == sample.text


@integration
class TestCustom:

    """Integration tests for custom attribute types."""

    def test_decorator(self, tmpdir):
        """Verify custom attribute types dump/load correctly."""
        tmpdir.chdir()
        sample = SampleCustomDecorated('sample')

        # check defaults
        assert '1.0' == sample.level

        # change values
        sample.level = '1.2.3'

        # check file values
        with open(sample.yorm_path, 'r') as stream:
            text = stream.read()
        assert """
        level: 1.2.3
        """.strip().replace("        ", "") + '\n' == text

        # change file values
        refresh_file_modification_times()
        text = """
        level: 1
        """.strip().replace("        ", "") + '\n'
        with open(sample.yorm_path, 'w') as stream:
            stream.write(text)

        # check object values
        assert '1' == sample.level


@integration
class TestInit:

    """Integration tests for initializing mapped classes."""

    def test_fetch_from_existing(self, tmpdir):
        """Verify attributes are updated from an existing file."""
        tmpdir.chdir()
        sample = SampleStandardDecorated('sample')
        sample2 = SampleStandardDecorated('sample')
        assert sample2.yorm_path == sample.yorm_path

        refresh_file_modification_times()

        logging.info("changing values in object 1...")
        sample.object = {'key2': 'value'}
        sample.array = [0, 1, 2]
        sample.string = "Hello, world!"
        sample.number_int = 42
        sample.number_real = 4.2
        sample.true = True
        sample.false = False

        logging.info("reading changed values in object 2...")
        assert 'value' == sample2.object.get('key2')
        assert [0, 1, 2] == sample2.array
        assert "Hello, world!" == sample2.string
        assert 42 == sample2.number_int
        assert 4.2 == sample2.number_real
        assert True is sample2.true
        assert False is sample2.false


@integration
class TestDelete:

    """Integration tests for deleting files."""

    def test_read(self, tmpdir):
        """Verify a deleted file cannot be read from."""
        tmpdir.chdir()
        sample = SampleStandardDecorated('sample')
        sample.yorm_mapper.delete()

        with pytest.raises(FileNotFoundError):
            print(sample.string)

        with pytest.raises(FileNotFoundError):
            sample.string = "def456"

    def test_write(self, tmpdir):
        """Verify a deleted file cannot be written to."""
        tmpdir.chdir()
        sample = SampleStandardDecorated('sample')
        sample.yorm_mapper.delete()

        with pytest.raises(FileNotFoundError):
            sample.string = "def456"

    def test_multiple(self, tmpdir):
        """Verify a deleted file can be deleted again."""
        tmpdir.chdir()
        sample = SampleStandardDecorated('sample')
        sample.yorm_mapper.delete()
        sample.yorm_mapper.delete()


if __name__ == '__main__':
    pytest.main()



#!/usr/bin/env python

"""Setup script for YORM."""

import setuptools

from yorm import __project__, __version__

import os
if os.path.exists('README.rst'):
    README = open('README.rst').read()
else:
    README = ""  # a placeholder, readme is generated on release
CHANGES = open('CHANGES.md').read()


setuptools.setup(
    name=__project__,
    version=__version__,

    description="Automatic object YAML mapping for Python.",
    url='https://github.com/jacebrowning/yorm',
    author='Jace Browning',
    author_email='jacebrowning@gmail.com',

    packages=setuptools.find_packages(),

    entry_points={'console_scripts': []},

    long_description=(README + '\n' + CHANGES),
    license='MIT',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Software Development :: Version Control',
        'Topic :: Text Editors :: Text Processing',
        'Topic :: System :: Filesystems',
    ],

    install_requires=open('requirements.txt').readlines(),
)



"""pytest configuration."""

import os
import pytest

ENV = 'TEST_INTEGRATION'  # environment variable to enable integration tests
REASON = "'{0}' variable not set".format(ENV)


def pytest_runtest_setup(item):
    """pytest setup."""
    if 'integration' in item.keywords and not os.getenv(ENV):
        pytest.skip(REASON)



"""Functions and decorators."""

import uuid

from . import common
from .base import Mappable
from .mapper import Mapper

log = common.logger(__name__)

UUID = 'UUID'


def sync(*args, **kwargs):
    """Convenience function to forward calls based on arguments.

    This function will call either:

    * `sync_object` - when given an unmapped object
    * `sync_instances` - when used as the class decorator

    Consult the signature of each call for more information.

    """
    if 'path_format' in kwargs or args and isinstance(args[0], str):
        return sync_instances(*args, **kwargs)
    else:
        return sync_object(*args, **kwargs)


def sync_object(instance, path, attrs=None, auto=True):
    """Enable YAML mapping on an object.

    :param instance: object to patch with YAML mapping behavior
    :param path: file path for dump/load
    :param attrs: dictionary of attribute names mapped to converter classes
    :param auto: automatically store attribute to file

    """
    attrs = attrs or {}
    _check_base(instance, mappable=False)

    class Mapped(Mappable, instance.__class__):

        """Original class with `Mappable` as the base."""

    instance.__class__ = Mapped

    instance.yorm_attrs = attrs or getattr(instance, 'yorm_attrs', attrs)
    instance.yorm_path = path
    instance.yorm_mapper = Mapper(instance.yorm_path)

    if not instance.yorm_mapper.exists:
        instance.yorm_mapper.create(instance)
        if auto:
            instance.yorm_mapper.store(instance, instance.yorm_attrs)
    else:
        instance.yorm_mapper.fetch(instance, instance.yorm_attrs)

    instance.yorm_mapper.auto = auto

    return instance


def sync_instances(path_format, format_spec=None, attrs=None, auto=True):
    """Class decorator to enable YAML mapping after instantiation.

    :param path_format: formatting string to create file paths for dump/load
    :param format_spec: dictionary to use for string formatting
    :param attrs: dictionary of attribute names mapped to converter classes
    :param auto: automatically store attribute to file

    """
    format_spec = format_spec or {}
    attrs = attrs or {}

    def decorator(cls):
        """Class decorator."""
        if hasattr(cls, 'yorm_attrs'):
            cls.yorm_attrs.update(attrs)
        else:
            cls.yorm_attrs = attrs

        class Mapped(Mappable, cls):

            """Original class with `Mappable` as the base."""

            def __init__(self, *_args, **_kwargs):
                super().__init__(*_args, **_kwargs)

                format_spec2 = {}
                for key, value in format_spec.items():
                    format_spec2[key] = getattr(self, value)
                if '{' + UUID + '}' in path_format:
                    format_spec2[UUID] = uuid.uuid4().hex
                format_spec2['self'] = self

                self.yorm_path = path_format.format(**format_spec2)
                self.yorm_mapper = Mapper(self.yorm_path)

                if not self.yorm_mapper.exists:
                    self.yorm_mapper.create(self)
                    if auto:
                        self.yorm_mapper.store(self, self.yorm_attrs)
                else:
                    self.yorm_mapper.fetch(self, self.yorm_attrs)

                self.yorm_mapper.auto = auto

        return Mapped

    return decorator


def attr(**kwargs):
    """Class decorator to map attributes to converters.

    :param kwargs: keyword arguments mapping attribute name to converter class

    """
    def decorator(cls):
        """Class decorator."""
        if not hasattr(cls, 'yorm_attrs'):
            cls.yorm_attrs = {}
        for name, converter in kwargs.items():
            cls.yorm_attrs[name] = converter
        return cls

    return decorator


def update(instance, fetch=True, force=True, store=True):
    """Synchronize changes between a mapped object and its file.

    :param instance: object with patched YAML mapping behavior
    :param fetch: update the object with changes from its file
    :param force: even if the file appears unchanged
    :param store: update the file with changes from the object

    """
    _check_base(instance, mappable=True)

    if fetch:
        update_object(instance, force=force)
    if store:
        update_file(instance)


def update_object(instance, force=True):
    """Synchronize changes into a mapped object from its file.

    :param instance: object with patched YAML mapping behavior
    :param force: update the object even if the file appears unchanged

    """
    _check_base(instance, mappable=True)

    instance.yorm_mapper.fetch(instance, instance.yorm_attrs, force=force)


def update_file(instance):
    """Synchronize changes into a mapped object's file.

    :param instance: object with patched YAML mapping behavior

    """
    _check_base(instance, mappable=True)

    instance.yorm_mapper.store(instance, instance.yorm_attrs)


def _check_base(obj, mappable=True):
    """Confirm an object's base class is `Mappable` as required."""
    if mappable and not isinstance(obj, Mappable):
        raise common.UseageError("{} is not mapped".format(repr(obj)))
    if not mappable and isinstance(obj, Mappable):
        raise common.UseageError("{} is already mapped".format(repr(obj)))
